#lang racket
;; Switch to an API that exposes events instead of
;; channels, because that will compose better.

(define-struct q (en-evt de-evt))

(define (create-queue)
  (define enqueue-ch (make-channel))
  (define dequeue-ch (make-channel))
  
  (define (serve items)
    (sync (handle-evt enqueue-ch
                      (lambda (item)
                        (serve (append items
                                       (list item)))))
          (if (empty? items)
              never-evt
              (handle-evt dequeue-ch
                          (lambda (reply-ch)
                            (channel-put reply-ch (first items))
                            (serve (rest items)))))))
  
  (thread (lambda () (serve empty)))
  
  (make-q (lambda (v)
            (channel-put-evt enqueue-ch v))
          (lambda (reply-ch)
            (channel-put-evt dequeue-ch reply-ch))))

(define q1 (create-queue))
(define q2 (create-queue))

(sync ((q-en-evt q1) 10))
(sync ((q-en-evt q2) 20))

(define (dequeue)
  (define reply-ch (make-channel))
  (sync ((q-de-evt q1) reply-ch)
        ((q-de-evt q2) reply-ch))
  (channel-get reply-ch))

(dequeue)
(dequeue)
