#lang racket
;; The additions of 12.rkt, but directly on 9.rkt:

;; A queue that accepts a predicate before supplying an
;; item from the queue. The server now needs to keep
;; a list of requests, and the synchronization point is
;; the request response, not the request acceptance.

(define-struct q (en-ch de-ch))

(define (create-queue)
  (define enqueue-ch (make-channel))
  (define dequeue-ch (make-channel))
  
  ;; A req (cons pred reply-ch)
  
  (define (serve items reqs)
    (apply
     sync 
     (handle-evt enqueue-ch
                 (lambda (item)
                   (serve (append items
                                  (list item))
                          reqs)))
     (handle-evt dequeue-ch
                 (lambda (req)
                   (serve items (cons req reqs))))
     (map (lambda (req)
            (match req
              [(cons pred reply-ch)
               (define ok-items (filter pred items))
               (if (empty? ok-items)
                   never-evt
                   (handle-evt (channel-put-evt reply-ch (first ok-items))
                               (lambda (_)
                                 (serve (remq (first ok-items) items)
                                        (remq req reqs)))))]))
           reqs)))

  (thread (lambda () (serve empty empty)))
  
  (make-q enqueue-ch dequeue-ch))

(define q1 (create-queue))

(channel-put (q-en-ch q1) 10)
(channel-put (q-en-ch q1) 21)

(define (dequeue q pred)
  (define reply-ch (make-channel))
  (channel-put (q-de-ch q) (cons pred reply-ch))
  (channel-get reply-ch))

(dequeue q1 odd?)
(dequeue q1 even?)
(thread (lambda () 
          (sleep 1) 
          (channel-put (q-en-ch q1) 30)
          (sleep 1) 
          (channel-put (q-en-ch q1) 41)))
(dequeue q1 odd?)
